"""
This test creates a real landscape deployment, and runs some checks against it.

FIXME: revert to using ssh -q, stderr=STDOUT instead of 2>&1, stderr=PIPE once
       lp:1281577 is addressed.
"""
import tempfile
from subprocess import check_output, PIPE

from helpers import IntegrationTest, StoppedService
from layers import OneLandscapeUnitLayer, OneLandscapeUnitNoCronLayer


class ServiceTest(IntegrationTest):

    layer = OneLandscapeUnitLayer

    def setUp(self):
        super(ServiceTest, self).setUp()
        self.frontend = self.environment.get_haproxy_public_address()

    def test_app(self):
        """Verify that the APP service is up.

        Specifically that it is reachable and that it presents the new
        user form.

        Note: In order to work on a new server or a server with the
        first admin user already created, this phrase should match
        the new-standalone-user form, the login form, and not
        the maintenance page.
        """
        self.environment.check_service("appserver")

    def test_msg(self):
        """Verify that the MSG service is up.

        Specifically that it is reachable and that it responds
        correctly to requests.
        """
        self.environment.check_service("msgserver")

    def test_ping(self):
        """Verify that the PING service is up.

        Specifically that it is reachable and that it responds
        correctly to a ping request without an ID.
        """
        self.environment.check_service("pingserver")

    def test_api(self):
        """Verify that the API service is up.

        Specifically that it is reachable and returns its name.
        """
        self.environment.check_service("api")

    def test_api_endpoint(self):
        """Verify that API endpoint is correctly listed."""
        self.environment.bootstrap_landscape(
            admin_name="foo", admin_password="bar", admin_email="foo@bar")

        cookie_jar = tempfile.NamedTemporaryFile()
        cookie_file = cookie_jar.name
        self.environment.login("foo@bar", "bar", cookie_file)

        public_url = self.environment.get_haproxy_public_address()
        self.environment.check_url(
            "/settings", "https://%s/api/" % public_url,
            cookie_jar=cookie_file)

    def test_upload(self):
        """Verify that the PACKAGE UPLOAD service is up.

        Specifically that it is reachable and returns its name.
        """
        self.environment.check_service("package-upload")

    def test_no_broker_defaults(self):
        """Verify that [broker] has no default values.

        This test verifies that the host and password configuration keys
        from the [broker] section don't remain at their default values.
        """
        config = self.environment.get_landscape_config()
        broker = config["broker"]
        self.assertNotEqual(broker["host"], "localhost")
        self.assertNotEqual(broker["password"], "landscape")

    def test_app_unavailable_page(self):
        """
        Verify that the frontend shows the styled unavailable page.
        """
        self.useFixture(
            StoppedService(self.environment, "landscape-appserver"))
        self.environment.check_url("/", "Service will resume shortly")

    def test_msg_unavailable_page(self):
        """
        Verify that the frontend shows the unavailable page for msg.
        """
        self.useFixture(
            StoppedService(self.environment, "landscape-msgserver"))
        self.environment.check_url(
            "/message-system", "Landscape is unavailable")

    def test_ping_unavailable_page(self):
        """
        Verify that the frontend shows the unavailable page for ping.
        """
        self.useFixture(
            StoppedService(self.environment, "landscape-pingserver"))
        self.environment.check_url(
            "/ping", "Landscape is unavailable", proto="http")

    def test_error_pages(self):
        """
        Verify that the offline pages from Landscape's offline dir got
        sent to haproxy and configured to used as error pages for HTTP
        status codes 403, 500, 502, 503 and 504.
        """
        # Look at the files on disk, since we can't instrument the
        # server to return the required HTTP codes or cause timeouts.
        error_files = {
            "403": "unauthorized-haproxy.html",
            "500": "exception-haproxy.html",
            "502": "unplanned-offline-haproxy.html",
            "503": "unplanned-offline-haproxy.html",
            "504": "timeout-haproxy.html"}
        haproxy_path = "/var/lib/haproxy/service_landscape-https/"
        ls_path = "/opt/canonical/landscape/canonical/landscape/offline/"
        for http_code, file_name in error_files.items():
            haproxy_file = self.environment.get_binary_file(
                haproxy_path + http_code + ".http", "haproxy")
            landscape_file = self.environment.get_binary_file(
                ls_path + file_name, "landscape-server")
            self.assertEqual(haproxy_file, landscape_file)

    def test_ssl_certificate_is_in_place(self):
        """
        The landscape-server charm looks at the SSL certificate set on the
        relation with haproxy and writes it on disk in the location that
        the application expects (it will need it when generating client
        configuration for Autopilot deployments).
        """
        ssl_cert = self.environment.get_text_file(
            "/etc/ssl/certs/landscape_server_ca.crt", "landscape-server")
        self.assertTrue(ssl_cert.startswith("-----BEGIN CERTIFICATE-----"))

    def test_landscape_packages_are_held(self):
        """
        Check that landscape packages are marked as "held" by apt-mark.
        """
        out, _ = self.environment.run_command_on_landscape("apt-mark showhold")
        self.assertIn("landscape-server", out)
        self.assertIn("landscape-hashid", out)


class CronTest(IntegrationTest):
    """Host all the tests that expects the cron daemon to be stopped.

    The deployment will the same minimal one from OneLandscapeUnitTest, but
    the cron daemon will be stopped, so Landscape cron jobs in particular
    won't be run.
    """
    layer = OneLandscapeUnitNoCronLayer

    def assertCronStatus(self, status, output, cron_name):
        message = "{}: output:{} status:{}".format(cron_name, output, status)
        self.assertEqual(status, 0, message)
        self.assertEqual(output, "", message)

    def test_maintenance_cron(self):
        """Verify that the maintenance cron job runs without errors."""
        script = "/opt/canonical/landscape/scripts/maintenance.sh"
        output, status = self.environment.run_script_on_cron_unit(
            script, self.layer)
        self.assertCronStatus(status, output, "maintenance_cron")

    def test_update_security_db_cron(self):
        """Verify that the update_security_db cron job runs without errors."""
        script = "/opt/canonical/landscape/scripts/update_security_db.sh"
        output, status = self.environment.run_script_on_cron_unit(
            script, self.layer)
        self.assertCronStatus(status, output, "update_security_db_cron")

    def test_update_alerts_cron(self):
        """Verify that the update_alerts cron job runs without errors."""
        script = "/opt/canonical/landscape/scripts/update_alerts.sh"
        output, status = self.environment.run_script_on_cron_unit(
            script, self.layer)
        self.assertCronStatus(status, output, "update_alerts_cron")

    def test_landscape_profiles_cron(self):
        """Verify that the landscape_profiles cron job runs without errors."""

        # process_profiles renamed to landscape_profiles on trunk @ r8238
        find_cmd = (
            "sudo ls /opt/canonical/landscape/scripts/landscape_profiles.sh"
            " || sudo ls /opt/canonical/landscape/scripts/process_profiles.sh")
        cmd = ["juju", "run", "--unit", self.layer.cron_unit, find_cmd]
        script = check_output(cmd, stderr=PIPE).decode("utf-8").strip()

        output, status = self.environment.run_script_on_cron_unit(
            script, self.layer)
        self.assertCronStatus(status, output, "profiles_cron")

    def test_process_alerts_cron(self):
        """Verify that the process_alerts cron job runs without errors."""
        script = "/opt/canonical/landscape/scripts/process_alerts.sh"
        output, status = self.environment.run_script_on_cron_unit(
            script, self.layer)
        self.assertCronStatus(status, output, "process_alerts_cron")

    def test_hash_id_databases_cron(self):
        """Verify that the hash_id_databases cron job runs without errors."""

        # Use the sample hashids config file, to avoid pulling APT lists
        # for the whole Ubuntu archive.
        config_dir = "/opt/canonical/landscape/configs/standalone"
        self.environment.run_command_on_landscape(
            "sudo cp {config_dir}/hash-id-databases-sample.conf"
            " {config_dir}/hash-id-databases.conf".format(
                config_dir=config_dir))

        script = "/opt/canonical/landscape/scripts/hash_id_databases.sh"
        output, status = self.environment.run_script_on_cron_unit(
            script, self.layer)
        self.assertCronStatus(status, output, "hash_id_databases_cron")

    def test_meta_releases_cron(self):
        """Verify that the meta_releases cron job runs without errors."""
        script = "/opt/canonical/landscape/scripts/meta_releases.sh"
        output, status = self.environment.run_script_on_cron_unit(
            script, self.layer)
        self.assertCronStatus(status, output, "meta_releases_cron")

    def test_sync_lds_releases_cron(self):
        """Verify that the sync_lds_releases cron job runs without errors."""
        script = "/opt/canonical/landscape/scripts/sync_lds_releases.sh"
        output, status = self.environment.run_script_on_cron_unit(
            script, self.layer)
        self.assertCronStatus(status, output, "sync_lds_releases_cron")

    def test_root_url_is_set(self):
        """
        The root URL should be set in service.conf.
        """
        config = self.environment.get_landscape_config()
        frontend = self.environment.get_haproxy_public_address()
        self.assertEqual(
            "https://%s/" % frontend, config["global"]["root-url"])
